#include <allegro.h>
#include "robotdat.h"
#include "robots.h"

#define puts(x) while(0)

DATAFILE *data;
BITMAP *buffer;

player_t *player;

game_object_t *objlist_head;

int cheat = 0;
int roundlost = 0;
int game_over = 0;

void add_object (game_object_t *obj)
{
   obj->next = objlist_head;
   objlist_head = obj;
}

void draw_object (game_object_t *obj)
{
   rotate_sprite(buffer, (BITMAP*)data[obj->graphic].dat, obj->x, obj->y, obj->angle);
}

int check_done (void)
{
   game_object_t *obj;

   obj = objlist_head;

   while (obj)
   {
      if (obj->type == TYPE_ROBOT && obj->state == ROBOT_LIVING)
         return 0;

      obj = obj->next;
   }

   return 1;
}

void move_obj (game_object_t *obj, int vel)
{
   fixed x, y;

   x = fmul (fcos (obj->angle - itofix(64)), itofix (vel));
   y = fmul (fsin (obj->angle - itofix(64)), itofix (vel));

   obj->x += fixtoi (x);
   obj->y += fixtoi (y);
}

void set_starting_pos (game_object_t *obj)
{
   obj->x = rand()%FIELD_W + FIELD_X;
   obj->y = rand()%FIELD_H + FIELD_Y;
}


void create_plane (void)
{
   plane_t *theplane;

   theplane = (plane_t*)malloc (sizeof (plane_t));

   get_random_xy(&(theplane->obj.x),&(theplane->obj.y));   
 
   theplane->obj.type = TYPE_PLANE;
   theplane->obj.state = PLANE_ENTERING;
   theplane->obj.graphic = PLANE;

   theplane->dest_x = mouse_x;
   theplane->dest_y = mouse_y;

   theplane->obj.angle = calculate_angle (theplane->obj.x, theplane->obj.y, theplane->dest_x, theplane->dest_y);

   add_object ((game_object_t*)theplane);
}

void do_ui (game_object_t *obj)
{
   static char mousepressed = 0;
   game_object_t *tmpobj;
   
   if (key[KEY_ESC])
      game_over = 1;

   if (!mouse_b & 1)
      mousepressed = 0;

   switch (obj->state)
   {
      case PLAYER_TELEPORT:
         if (obj->graphic == TELE7)
         {
            set_starting_pos (obj);
            obj->state = PLAYER_LIVING;
            obj->graphic = PLAYER;
         }
         else
            obj->graphic ++;

         break;

      case PLAYER_LIVING:
         if (key[KEY_LEFT])
            obj->angle -= itofix (8);

         if (key[KEY_RIGHT])
            obj->angle += itofix (8);

         if (key[KEY_UP])
            move_obj (obj, 5);

         if (key[KEY_DOWN])
            move_obj (obj, -3);

         if (key[KEY_T])
         {
            if (player->teleports > 0)
            {
               player->teleports --;
               obj->state = PLAYER_TELEPORT;
               obj->graphic = TELE1;
            }
         }

         if (mouse_b & 1 && mousepressed == 0)
         {
            if (player->airstrikes > 0)
            {
               player->airstrikes --;
               create_plane();
            }
            mousepressed = 1;
         }

         tmpobj = objlist_head;

         while (tmpobj)
         {
            if (tmpobj->type == TYPE_JUNK)
               if (checkcollision (obj, tmpobj))
                  move_obj (obj, -5); // bounce back when hitting junk
            tmpobj = tmpobj->next;
         }
         break;
      case PLAYER_DYING:
         if (obj->graphic == EXPLO3)
         {
            obj->state = PLAYER_DEAD;
            rest(500);
            roundlost = 1;
         }
         else
            obj->graphic ++;
         break;
      case PLAYER_DEAD: // No time to implement more graphics!
         break;
   }
}

int checkcollision (game_object_t *obj1, game_object_t *obj2)
{
   struct rect
   {
      int x1, x2;
      int y1, y2;
   };

   struct rect r1, r2;

   r1.x1 = obj1->x;
   r1.y1 = obj1->y;
   r1.x2 = obj1->x + 32;
   r1.y2 = obj1->y + 32;

   r2.x1 = obj2->x;
   r2.y1 = obj2->y;
   r2.x2 = obj2->x + 32;
   r2.y2 = obj2->y + 32;

   if ((r1.x1 >= r2.x1 && r1.x1 <= r2.x2 && r1.y1 >= r2.y1 && r1.y1 <= r2.y2) ||
       (r1.x1 >= r2.x1 && r1.x1 <= r2.x2 && r1.y2 >= r2.y1 && r1.y2 <= r2.y2) ||
       (r1.x2 >= r2.x1 && r1.x2 <= r2.x2 && r1.y1 >= r2.y1 && r1.y1 <= r2.y2) ||
       (r1.x2 >= r2.x1 && r1.x2 <= r2.x2 && r1.y2 >= r2.y1 && r1.y2 <= r2.y2) )
      return 1;

   return 0;
}

void do_robot (game_object_t *robot)
{
   game_object_t *obj;

   switch (robot->state)
   {
      case ROBOT_LIVING:
         robot->angle = calculate_angle( robot->x, robot->y, player->obj.x, player->obj.y);
         move_obj (robot, 1);

         obj = objlist_head;

         while (obj)
         {
            if (obj == robot)
            {
               obj = obj->next;
               continue;
            }

            if (checkcollision (obj, robot))
            {
               if (obj->type == TYPE_PLAYER && obj->state == PLAYER_LIVING)
               {
                  player->obj.state = PLAYER_DYING;
                  player->obj.graphic = EXPLO1;
               }

               if (obj->type == TYPE_ROBOT)
               {
                  obj->state = robot->state = ROBOT_DYING;
                  obj->graphic = robot->graphic = EXPLO1;
                  obj->angle = robot->angle = itofix(0);
               }

               if (obj->type == TYPE_JUNK)
               {
                  robot->state = ROBOT_DYING;
                  robot->graphic = EXPLO1;
                  robot->angle = itofix(0);
               }
            }
            obj = obj->next;
         }
         break;
      case ROBOT_DYING:
         if (robot->graphic == EXPLO3)
         {
            robot->type = TYPE_JUNK;
            robot->graphic = JUNK1;
            robot->state = JUNK;
         }
         else
            robot->graphic ++;
         break;
   }
}

void do_plane(plane_t *plane)
{
   int index;

   switch (plane->obj.state)
   {
      case PLANE_ENTERING:
         move_obj ((game_object_t*)plane, 15);
         if (abs(
               calculate_angle (plane->obj.x, plane->obj.y, plane->dest_x, plane->dest_y)
               - plane->obj.angle) > itofix(10))
         {
            // drop a bomb!
            for (index = 0; index < 5; index ++)
               create_junk ((rand() % 50) + plane->dest_x - 25, (rand() % 50) + plane->dest_y - 25);

            plane->obj.state = PLANE_LEAVING;
            get_random_xy (&(plane->dest_x), &(plane->dest_y));
            plane->obj.angle = calculate_angle (plane->obj.x, plane->obj.y, plane->dest_x, plane->dest_y);
         }
         break;
      case PLANE_LEAVING:
         move_obj ((game_object_t*)plane, 8);
         if (abs(
               calculate_angle (plane->obj.x, plane->obj.y, plane->dest_x, plane->dest_y)
               - plane->obj.angle) > itofix(10))
            destroy_plane();
         break;
   }
}

void do_junk(game_object_t *obj)
{
   if (obj->graphic == JUNK3)
      obj->graphic = JUNK1;
   else
      obj->graphic ++;

}

void do_game (void)
{
   game_object_t *tmpobj;

   tmpobj = objlist_head;

   while (tmpobj)
   {
      switch (tmpobj->type)
      {
         case TYPE_PLAYER:
            do_ui (tmpobj);
            break;
         case TYPE_ROBOT:
            do_robot (tmpobj);
            break;
         case TYPE_PLANE:
            do_plane ((plane_t*)tmpobj);
            break;
         case TYPE_JUNK:
            do_junk (tmpobj);
            break;
      }

      draw_object (tmpobj);
      tmpobj = tmpobj->next;
   }

   draw_buffer();
}

void get_random_xy(int *x, int *y)
{
   unsigned char orientation;

   orientation = rand() % 4;

   switch (orientation)
   {
      case 0:
         *x = -50;
         *y = rand() % 480;
         break;
      case 1:
         *x = 700;
         *y = rand() % 480;
         break;
      case 2:
         *y = -50;
         *x = rand() % 640;
         break;
      case 3:
         *y = 550;
         *x = rand() % 640;
         break;
   }
}

fixed calculate_angle (int x, int y, int dest_x, int dest_y)
{
   fixed angle;
   unsigned char quad;

   if (dest_x == x && dest_y > y)
      return (itofix (128));
   if (dest_x == x && dest_y < y)
      return (itofix (0));
   if (dest_y == y && dest_x > x)
      return (itofix (64));
   if (dest_y == y && dest_x < x)
      return (itofix (196));


/*   if (dest_x > x && dest_y > y)
      quad = 1;
   if (dest_x < x && dest_y > y)
      quad = 2;
   if (dest_x < x && dest_y < y)
      quad = 3;
   if (dest_x < x && dest_y < y)
      quad = 4;
*/
   angle = fatan (fdiv (itofix (dest_y - y), itofix (dest_x - x)));

   angle -= itofix (64);

   if (dest_x > x)
      angle -= itofix (128);

   return (angle);
}


void intro(void)
{
   set_palette (black_palette);
   draw_rle_sprite (screen, (RLE_SPRITE*)data[BEAKER_LOGO].dat, 0, 60);
   fade_in ((RGB*)data[GAME_PAL].dat, 2);
   rest (2000);
   fade_out(2);
   clear(screen);
   set_palette ((RGB*)data[GAME_PAL].dat);
}

void create_buffer(void)
{
   buffer = create_bitmap (640, 480);
   clear (buffer);
}

void create_player (void)
{
   player = (player_t*)malloc (sizeof (player_t));

   set_starting_pos ((game_object_t*)player);

   player->obj.type = TYPE_PLAYER;
   player->obj.state = PLAYER_LIVING;
   player->obj.graphic = PLAYER;
   player->obj.angle = itofix (0);
   if (cheat)
   {
      player->teleports = 999;
      player->airstrikes = 999;
      player->score     = -10000;
   }
   else
   {
      player->teleports = 10;
      player->airstrikes = 2;
      player->score = 0;
   }
   add_object ((game_object_t*)player);
}

void destroy_plane (void)
{
   game_object_t *obj;
   game_object_t *prevobj;

   prevobj = NULL;
   obj = objlist_head;
   while (obj)
   {
      if (obj->type == TYPE_PLANE)
      {
         if (prevobj)
            prevobj->next = obj->next;
         else
            objlist_head = obj->next;

         free (obj);
         return;
      }
      prevobj = obj;
      obj = obj->next;
   }
}

void create_robot (void)
{
   game_object_t *robot;

   robot = (game_object_t*)malloc (sizeof (game_object_t));

   set_starting_pos (robot);

   robot->type = TYPE_ROBOT;
   robot->state = ROBOT_LIVING;
   robot->graphic = ROBOT;
   robot->angle = calculate_angle( robot->x, robot->y, player->obj.x, player->obj.y);

   add_object (robot);
}

void create_junk (int x, int y)
{
   game_object_t *junk;

   junk = (game_object_t*)malloc (sizeof (game_object_t));

   if (x == POS_RAND && y == POS_RAND)
      set_starting_pos (junk);
   else
   {
      junk->x = x;
      junk->y = y;
   }

   junk->type = TYPE_JUNK;
   junk->state = JUNK;
   junk->graphic = JUNK1;
   junk->angle = itofix(0);

   add_object (junk);
}

void create_objects(int robots, int junk)
{
   int index;

   create_player ();

   for (index = 0; index < robots; index ++)
      create_robot();
   for (index = 0; index < junk; index ++)
      create_junk(POS_RAND, POS_RAND);
}

void destroy_objects (void)
{
   game_object_t *obj;
   game_object_t *nextobj;

   obj = NULL;
   nextobj = objlist_head;

   while (nextobj)
   {
      obj = nextobj;
      nextobj = obj->next;
      free (obj);
   }

   objlist_head = NULL;
}

void draw_buffer(void)
{
   show_mouse (buffer);
   blit (buffer, screen, 0, 0, 0, 0, 640, 480);
   show_mouse (NULL);
   clear (buffer);
}

void gameloop (void)
{
   int robots;
   int junk;

   robots = 5;
   junk = 5;

   while (!game_over)
   {
      do_game();
      rest (50);

      if (roundlost)
      {
         roundlost = 0;
         destroy_objects();
         create_objects (robots, junk);
      }

      if (check_done())
      {
         // round is over!
         destroy_objects();
         robots += 3;
         junk += 2;
         create_objects (robots, junk);
      }
      
   }

   destroy_objects();
   
}

int main (int argc, char *argv[])
{
   allegro_init();
   install_timer();
   install_keyboard();
   install_mouse();

   srand(time(0));
   puts ("Robots: The Next Generation");
   puts ("------------------------------ - - -  -   -       -");

   if (argc == 2 && strcmp (argv[1], "itsabeakertodd") == 0)
   {
      cheat = 1;
      puts ("Cheat mode enabled!  Muahahaha.");
   }
   puts ("Loading datafile...");
   data = load_datafile ("robots.dat");
   puts ("Creating buffer...");
   create_buffer();
   puts ("Creating life...");
   create_objects(5, 5);
   puts ("Reticulating splines...");
   rest (500);

   fade_out (2);

   if (set_gfx_mode (GFX_AUTODETECT_WINDOWED, 640, 480, 0, 0) < 0)
   {
      puts ("Unable to switch video card.  You suck.");
      exit (-1);
   }

   intro();

   set_mouse_sprite ((BITMAP*)data[TARGET].dat);
   set_mouse_sprite_focus (15, 15);

   gameloop();

   fade_out (2);

   set_gfx_mode (GFX_TEXT, 80, 25, 0, 0);
   puts ("Coded 06/27/00 by Jeremy Penner in 4 hours.");
}
END_OF_MAIN()
